"""Calculate the power density spectrum and other plasma parameters."""

from typing import Callable, Optional, Tuple

import numpy as np
import scipy.constants as const

from inscar import config, integrand_functions, numba_integration


class SpectrumCalculation:
    """Class containing the calculation of the power density spectrum."""

    def __init__(self):
        """Create the basic spectrum calculation object with empty attributes."""
        self.ion: config.Particle
        self.electron: config.Particle
        self.ion_integration_function: integrand_functions.Integrand
        self.electron_integration_function: integrand_functions.Integrand
        self._calulate_f = self._calulate_f_function
        self._susceptibility = self._susceptibility_function
        self.params: config.Parameters

    def set_params(self, params: config.Parameters) -> None:
        """Set the plasma parameters.

        Parameters
        ----------
        params : Parameters
            An instance of the ``Parameters`` class.
        """
        self.params = params

    def set_ion(self, ion: config.Particle) -> None:
        """Set the ion particles to use.

        Parameters
        ----------
        ion : Particle
            An instance of the ``Particle`` class, representing ions.
        """
        self.ion = ion

    def set_electron(self, electron: config.Particle) -> None:
        """Set the electron particles to use.

        Parameters
        ----------
        electron : Particle
            An instance of the ``Particle`` class, representing electrons.
        """
        self.electron = electron

    def set_ion_integration_function(
        self, function: integrand_functions.Integrand
    ) -> None:
        """Set the ion integration method to use.

        Parameters
        ----------
        function : Integrand
            An object of type ``Integrand``, representing the ions.
        """
        self.ion_integration_function = function

    def set_electron_integration_function(
        self, function: integrand_functions.Integrand
    ) -> None:
        """Set the electron integration method to use.

        Parameters
        ----------
        function : Integrand
            An object of type ``Integrand``, representing the electrons.
        """
        self.electron_integration_function = function

    def calculate_spectrum(self) -> Tuple[np.ndarray, np.ndarray]:
        """Calculate a spectrum based on the given parameters and calculation methods.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The frequency axis in the first position and the spectrum values along the
            axis in the second position.

        Raises
        ------
        ValueError
            If any of the needed attribute objects are not set.
        """
        if not hasattr(self, "params"):
            raise ValueError("No parameters set. Use set_params().")
        if not hasattr(self, "ion"):
            raise ValueError("No ion particle set. Use set_ion().")
        if not hasattr(self, "electron"):
            raise ValueError("No electron particle set. Use set_electron().")
        if not hasattr(self, "ion_integration_function"):
            raise ValueError(
                "No ion integration function set. Use set_ion_integration_function()."
            )
        if not hasattr(self, "electron_integration_function"):
            raise ValueError(
                "No electron integration function set. "
                + "Use set_electron_integration_function()."
            )

        fi = self._calulate_f(self.ion, self.ion_integration_function)
        fe = self._calulate_f(self.electron, self.electron_integration_function)

        xp_i = self._susceptibility(self.ion, self.ion_integration_function)
        xp_e = self._susceptibility(self.electron, self.electron_integration_function)

        with np.errstate(divide="ignore", invalid="ignore"):
            numerator1 = np.imag(-fe) * np.abs(1 + 2 * xp_i**2 * fi) ** 2
            numerator2 = 4 * xp_e**4 * np.imag(-fi) * np.abs(fe) ** 2
            numerator = numerator1 + numerator2
            denominator = np.abs(1 + 2 * xp_e**2 * fe + 2 * xp_i**2 * fi) ** 2
            spectrum = (
                self.electron.number_density
                / (np.pi * self.params.angular_frequency)
                * numerator
                / denominator
            )
        return self.params.linear_frequency, spectrum

    def set_calculate_f_function(
        self, f_func: Callable[[config.Particle, integrand_functions.Integrand], float]
    ) -> None:
        """Set what function to use to calculate the :math:`F` function.

        Parameters
        ----------
        f_func : Callable[[Particle, Integrand], float]
            A function that take a particle and an integrand function as input, and that
            calculates the :math:`F` function based on these, returning a numpy array.
            By default, the ``integrate`` function from the ``numba_integration`` module
            is used.

        See Also
        --------
        inscar.numba_integration.integrate
        """
        self._calulate_f = f_func

    def _calulate_f_function(
        self, particle: config.Particle, int_func: integrand_functions.Integrand
    ) -> np.ndarray:
        int_func.initialize(self.params, particle)
        the_type = int_func.the_type
        integrand = int_func.integrand()
        characteristic_velocity = getattr(int_func, "char_vel", None)
        return numba_integration.integrate(
            self.params,
            particle,
            integrand,
            the_type,
            characteristic_velocity,
        )

    def set_susceptibility_function(
        self, func: Callable[[config.Particle, integrand_functions.Integrand], float]
    ) -> None:
        """Set what function to use to calculate the susceptibility function.

        Parameters
        ----------
        func : Callable[[Particle, Integrand], float]
            A function that take a particle and an integrand function as input, and that
            calculates the susceptibility function based on these, returning a single
            float.
        """
        self._susceptibility = func

    def _susceptibility_function(
        self, particle: config.Particle, int_func: integrand_functions.Integrand
    ) -> float:
        kappa = getattr(particle, "kappa", 1)
        temp = particle.temperature
        if int_func.the_type == "maxwell":
            debye_length = get_debye_length(particle.number_density, temp)
            xp = np.sqrt(1 / (2 * debye_length**2 * self.params.radar_wavenumber**2))
        elif int_func.the_type == "kappa":
            debye_length = get_debye_length(particle.number_density, temp, kappa=kappa)
            xp = np.sqrt(1 / (2 * debye_length**2 * self.params.radar_wavenumber**2))
        elif int_func.the_type == "a_vdf":
            char_vel = getattr(int_func, "char_vel", 1)
            debye_length = get_debye_length(
                particle.number_density, temp, char_vel=char_vel
            )
            xp = np.sqrt(1 / (2 * debye_length**2 * self.params.radar_wavenumber**2))
        else:
            raise ValueError("Unknown function type.")
        return xp


def get_debye_length(
    number_density: float,
    electron_temperature: float,
    ion_temperature: Optional[float] = None,
    kappa: Optional[float] = None,
    char_vel: Optional[float] = None,
) -> float:
    """Calculate the Debye length.

    Parameters
    ----------
    number_density : float
        The number density of the plasma.
    electron_temperature : float
        The electron temperature.
    ion_temperature : float, optional
        The ion temperature.
    kappa : float, optional
        Kappa parameter.
    char_vel : float, optional
        Characteristic velocity.

    Returns
    -------
    float : float
        Debye length.
    """
    vacuum_permittivity = 1e-09 / 36 / np.pi

    if ion_temperature is None:
        if kappa is not None:
            length = np.sqrt(
                vacuum_permittivity
                * const.k
                * electron_temperature
                / (max(0, number_density) * const.e**2)
            ) * np.sqrt((kappa - 3 / 2) / (kappa - 1 / 2))
        elif char_vel is not None:
            length = np.sqrt(
                vacuum_permittivity
                * const.k
                * electron_temperature
                / (max(0, number_density) * const.e**2)
            ) * np.sqrt(char_vel)
        else:
            length = np.sqrt(
                vacuum_permittivity
                * const.k
                * electron_temperature
                / (max(0, number_density) * const.e**2)
            )
    else:
        length = np.sqrt(
            vacuum_permittivity
            * const.k
            / (
                (
                    max(0, number_density) / electron_temperature
                    + max(0, number_density) / ion_temperature
                )
                / const.e**2
            )
        )

    return length



"""Script containing the integrands used in the Gordeyev integral."""

from abc import ABC, abstractmethod, abstractproperty

import numpy as np
import scipy.constants as const
import scipy.integrate as si
import scipy.special as sps

from inscar import config, numba_integration, vdfs


class Integrand(ABC):
    """Base class for an integrand object.

    Parameters
    ----------
    ABC : ABC
        Abstract base class
    """

    @abstractproperty
    def the_type(self) -> str:
        """Return the type of the integrand implementation."""
        ...

    @abstractmethod
    def initialize(self, params: config.Parameters, particle: config.Particle) -> None:
        """Initialise the kappa integrand object.

        Parameters
        ----------
        params : Parameters
            A `Parameters` object.
        particle : Particle
            A `Particle` object.
        """

    @abstractmethod
    def integrand(self) -> np.ndarray:
        """Return the `np.ndarray` that is used as the integrand."""


class IntKappa(Integrand):
    """Implementation of the integrand of the Gordeyev integral.

    Notes
    -----
    This implementation is for the kappa distribution described by Mace [1]_.

    References
    ----------
    .. [1] R. L. Mace, "A Gordeyev integral for electrostatic waves in a magnetized
       plasma with a kappa velocity distribution," Physics of plasmas, vol. 10, no. 6,
       pp. 2101-2193, 2003.
    """

    @property
    def the_type(self) -> str:
        """Return the type of the VDF that is being used."""
        return "kappa"

    def __init__(self) -> None:
        """Set the attributes of the class."""
        self.params: config.Parameters
        self.particle: config.Particle
        self.gyro_frequency: float
        self.Z: np.ndarray
        self.Kn: np.ndarray

    def initialize(self, params: config.Parameters, particle: config.Particle) -> None:
        """Initialize the integration."""
        self.params = params
        self.particle = particle
        self.gyro_frequency = (
            const.e * self.params.magnetic_field_strength / self.particle.mass
        )
        self._z_func()

    def _z_func(self) -> None:
        y = self.particle.gordeyev_axis
        theta_2 = (
            2
            * ((self.particle.kappa - 3 / 2) / self.particle.kappa)
            * self.particle.temperature
            * const.k
            / self.particle.mass
        )
        self.Z = (2 * self.particle.kappa) ** (1 / 2) * (
            self.params.radar_wavenumber**2
            * np.sin(self.params.aspect_angle) ** 2
            * theta_2
            / self.gyro_frequency**2
            * (1 - np.cos(self.gyro_frequency * y))
            + 1
            / 2
            * self.params.radar_wavenumber**2
            * np.cos(self.params.aspect_angle) ** 2
            * theta_2
            * y**2
        ) ** (1 / 2)
        self.Kn = sps.kv(self.particle.kappa + 1 / 2, self.Z)
        self.Kn[self.Kn == np.inf] = 1

    def integrand(self) -> np.ndarray:
        """Return the integrand that goes into the Gordeyev integral."""
        y = self.particle.gordeyev_axis
        return (
            self.Z ** (self.particle.kappa + 0.5)
            * self.Kn
            * np.exp(-y * self.particle.collision_frequency)
        )


class IntMaxwell(Integrand):
    """Implementation of the integrand of the Gordeyev integral.

    Notes
    -----
    This implementation is for the integral for the Maxwellian distribution from e.g.
    Hagfors [1]_ or Mace [2]_.

    References
    ----------
    .. [1] T. Hagfors, "Density Fluctuations in a Plasma in a Magnetic Field, with
       Applications to the Ionosphere," Journal of Geophysical Research, vol. 66, no.
       9, pp. 1699-1712, 1961.
    .. [2] R. L. Mace, "A Gordeyev integral for electrostatic waves in a magnetized
       plasma with a kappa velocity distribution," Physics of plasmas, vol. 10, no. 6,
       pp. 2101-2193, 2003.
    """

    @property
    def the_type(self) -> str:
        """Return the type of the VDF that is being used."""
        return "maxwell"

    def __init__(self) -> None:
        """Set the attributes of the class."""
        self.params: config.Parameters
        self.particle: config.Particle
        self.gyro_frequency: float

    def initialize(self, params: config.Parameters, particle: config.Particle) -> None:
        """Initialize the integration."""
        self.params = params
        self.particle = particle
        self.gyro_frequency = (
            const.e * self.params.magnetic_field_strength / self.particle.mass
        )

    def integrand(self) -> np.ndarray:
        """Return the integrand that goes into the Gordeyev integral."""
        return np.exp(
            -self.particle.gordeyev_axis * self.particle.collision_frequency
            - self.params.radar_wavenumber**2
            * np.sin(self.params.aspect_angle) ** 2
            * self.particle.temperature
            * const.k
            / (self.particle.mass * self.gyro_frequency**2)
            * (1 - np.cos(self.gyro_frequency * self.particle.gordeyev_axis))
            - 0.5
            * (
                self.params.radar_wavenumber
                * np.cos(self.params.aspect_angle)
                * self.particle.gordeyev_axis
            )
            ** 2
            * self.particle.temperature
            * const.k
            / self.particle.mass
        )


class IntLong(Integrand):
    """Implementation of the integrand of the Gordeyev integral.

    Notes
    -----
    This implementation is for the integral for the isotropic distribution from Mace
    [1]_.

    References
    ----------
    .. [1] R. L. Mace, "A Gordeyev integral for electrostatic waves in a magnetized
       plasma with a kappa velocity distribution," Physics of plasmas, vol. 10, no. 6,
       pp. 2101-2193, 2003.
    """

    @property
    def the_type(self) -> str:
        """Return the type of the VDF that is being used."""
        return "a_vdf"

    def __init__(self) -> None:
        """Set the attributes of the class."""
        self.params: config.Parameters
        self.particle: config.Particle
        self.char_vel: float
        self.vdf = vdfs.VdfMaxwell
        self.gyro_frequency: float

    def set_vdf(self, vdf) -> None:
        """Assign a new VDF that is used in the integration."""
        self.vdf = vdf

    def initialize(self, params: config.Parameters, particle: config.Particle) -> None:
        """Initialize the integration."""
        self.params = params
        self.particle = particle
        self.gyro_frequency = (
            const.e * self.params.magnetic_field_strength / self.particle.mass
        )

    def _v_int(self) -> np.ndarray:
        v = self.particle.velocity_axis
        y = self.particle.gordeyev_axis
        f = self.vdf(self.params, self.particle)

        # Compare the velocity integral to the Maxwellian case. This way we make up for
        # the change in characteristic velocity and Debye length for different particle
        # distributions.
        res_maxwell = numba_integration.integrate_velocity(
            self.particle.gordeyev_axis,
            v,
            vdfs.VdfMaxwell(self.params, self.particle).f_0(),
            self.params.radar_wavenumber,
            self.params.aspect_angle,
            self.gyro_frequency,
        )
        int_maxwell = si.simps(res_maxwell, y)
        v_func = f.f_0()
        res = numba_integration.integrate_velocity(
            self.particle.gordeyev_axis,
            v,
            v_func,
            self.params.radar_wavenumber,
            self.params.aspect_angle,
            self.gyro_frequency,
        )
        int_res = si.simps(res, y)
        # The scaling of the factor describing the characteristic velocity
        self.char_vel = int_maxwell / int_res
        print(
            f"Debye length of the current distribution is {self.char_vel}"
            + " times the Maxwellian Debye length."
        )
        return res

    def _p_d(self) -> np.ndarray:
        y = self.particle.gordeyev_axis
        # At $ y=0 $ we get $ 0/0 $, so we use
        # $ \lim_{y\rightarrow 0^+}\mathrm{d}p/\mathrm{d}y = |k| |w_c| / \sqrt(w_c^2) $
        # (from above, opposite sign from below)
        cos_t = np.cos(self.params.aspect_angle)
        sin_t = np.sin(self.params.aspect_angle)
        w_c = self.gyro_frequency
        num = (
            abs(self.params.radar_wavenumber)
            * abs(w_c)
            * (cos_t**2 * w_c * y + sin_t**2 * np.sin(w_c * y))
        )
        term1 = (cos_t * w_c * y) ** 2
        term2 = -2 * sin_t**2 * np.cos(w_c * y)
        term3 = 2 * sin_t**2
        den = w_c * (term1 + term2 + term3) ** 0.5
        # np.sign(y[-1]) takes care of weather the limit should be considered taken from
        # above or below. The last element of the np.ndarray is chosen since it is
        # assumed y runs from 0 to some finite real number.
        first = np.sign(y[-1]) * abs(self.params.radar_wavenumber) * abs(w_c) / abs(w_c)
        with np.errstate(divide="ignore", invalid="ignore"):
            out = num / den
        zero = 0.0
        out[np.where(den == zero)[0]] = first

        return out

    def integrand(self) -> np.ndarray:
        """Return the integrand that goes into the Gordeyev integral."""
        return self._p_d() * self._v_int()



"""Implementation of integrals using ``numba`` for parallelization."""

import math
from typing import Optional

import numba as nb
import numpy as np
import scipy.constants as const

from inscar import config


@nb.njit(parallel=True, cache=True)
def trapz(y, x):
    """Pure python version of trapezoid rule.

    Parameters
    ----------
    y : np.ndarray
        Function to integrate over.
    x : np.ndarray
        The axis to integrate along.

    Returns
    -------
    float
        The value of the integral.
    """
    s = 0
    for i in nb.prange(1, len(x)):
        s += (x[i] - x[i - 1]) * (y[i] + y[i - 1])
    return s / 2


@nb.njit(cache=True)
def inner_int(w, x, function):
    """Calculate the Gordeyev integral of the F function.

    Parameters
    ----------
    w : np.ndarray
        Angular frequency array.
    x : np.ndarray
        The axis of the :math:`F` function.
    function : np.ndarray
        Function to integrate over.

    Returns
    -------
    np.ndarray
        The integrated function.
    """
    array = np.zeros_like(w, dtype=np.complex128)
    for idx in nb.prange(len(w)):
        array[idx] = trapz(np.exp(-1j * w[idx] * x) * function, x)
    return array


def integrate(
    params: config.Parameters,
    particle: config.Particle,
    integrand: np.ndarray,
    the_type: str,
    char_vel: Optional[float] = None,
) -> np.ndarray:
    r"""Calculate the Gordeyev integral for each frequency.

    This locates the Gordeyev axis of the particle object, and for each frequency
    calculates the Gordeyev integral corresponding to it, returning an array of the same
    length as the frequency array.

    Parameters
    ----------
    params : Parameters
        ``Parameters`` object with simulation parameters.
    particle : Particle
        ``Particle`` object with particle parameters.
    integrand : np.ndarray
        Function to integrate over.
    the_type : str
        Defines which type of ``Integrand`` class that is used.
    char_vel : float, optional
        Characteristic velocity of the particle.

    Returns
    -------
    np.ndarray
        The integral evaluated along the whole Gordeyev frequency axis of the particle.

    See Also
    --------
    inscar.spectrum_calculation.SpectrumCalculation.set_calculate_f_function

    Notes
    -----
    The :math:`F` function here refers to the equation found in Hagfors [1]_:

    .. math::

       \\begin{aligned}
       F_{\\mathrm{e}}(\\boldsymbol{k}, \\omega)
       =1-\\left(i\\frac{X(\\omega)}{X_{\\mathrm{e}}}+\\Lambda_{\\mathrm{e}}\\right)\\int_0^\\infty\\exp\\left
       \\{-iy\\frac{X(\\omega)}{X_{\\mathrm{e}}}-y\\Lambda_{\\mathrm{e}}\\right.\\\
       \\left.-\\frac{1}{2X_{\\mathrm{e}}^2}\\left[\\sin^2\\theta(1-\\cos
       y)+\\frac{1}{2}y^2\\cos^2\\theta\\right]\\right \\}\\textnormal{d}y
       \\end{aligned}

    References
    ----------
    .. [1] T. Hagfors, "Density Fluctuations in a Plasma in a Magnetic Field, with
        Applications to the Ionosphere," Journal of Geophysical Research, vol. 66,
        no. 9, pp. 1699-1712, 1961.
    """
    y = particle.gordeyev_axis
    temp = particle.temperature
    mass = particle.mass
    w = params.angular_frequency
    nu = particle.collision_frequency
    array = inner_int(w, y, integrand)
    if the_type == "a_vdf":
        # Characteristic velocity scaling
        a = 4 * np.pi * temp * const.k * array / mass * char_vel
    elif the_type == "kappa":
        a = array / (2 ** (particle.kappa - 1 / 2) * math.gamma(particle.kappa + 1 / 2))
    else:
        a = array
    return a if the_type == "a_vdf" else 1 - (1j * w + nu) * a


@nb.njit(cache=True)
def integrate_velocity(y, v, f, k_r, theta, w_c):  # noqa: PLR0913
    """Calculate the velocity integral.

    Parameters
    ----------
    y : np.ndarray
        The axis in frequency space.
    v : np.ndarray
        The axis in velocity space.
    f : np.ndarray
        The function to integrate.
    k_r : float
        The radar wavenumber.
    theta : float
        The radar aspect angle.
    w_c : float
        The gyro frequency.

    Returns
    -------
    np.ndarray
        The integrated function.
    """
    array = np.zeros_like(y)
    for idx in nb.prange(len(y)):
        array[idx] = trapz(v * np.sin(p(y[idx], k_r, theta, w_c) * v) * f, v)
    return array


@nb.njit(cache=True)
def p(y, k_r, theta, w_c):
    """Calculate the :math:`p` function.

    Parameters
    ----------
    y : np.ndarray
        Parameter from Gordeyev integral
    k_r : float
        The radar wave number.
    theta : float
        The radar aspect angle.
    w_c : float
        The gyro frequency.

    Returns
    -------
    np.ndarray
        Value of the :math:`p` function

    Notes
    -----
    Implementation of the :math:`p` function from Mace [1]_.

    References
    ----------
    .. [1] R. L. Mace, "A Gordeyev integral for electrostatic waves in a magnetized
       plasma with a kappa velocity distribution," Physics of plasmas, vol. 10, no. 6,
       pp. 2101-2193, 2003.
    """
    k_perp = k_r * np.sin(theta)
    k_par = k_r * np.cos(theta)
    return (2 * k_perp**2 / w_c**2 * (1 - np.cos(y * w_c)) + k_par**2 * y**2) ** 0.5



"""Extra, experimental velocity distribution functions."""

import numpy as np
import scipy.constants as const
import scipy.integrate as si
import scipy.special as sps

import inscar as isr


class VdfKappa2(isr.Vdf):
    """Create an object that make kappa vol. 2 distribution functions.

    Notes
    -----
    Kappa VDF used in dispersion relation paper by Ziebell, Gaelzer and Simões
    [1]_. Defined by Leubner [2]_.

    References
    ----------
    .. [1] L. Ziebell, R. Gaelzer, & F. Simões, "Dispersion relation for electrostatic
       waves in plasmas with isotropic and anisotropic Kappa distributions for electrons
       and ions," Journal of Plasma Physics, vol. 83, no. 5, pp. 905830503, 2017.
       doi:10.1017/S0022377817000733
    .. [2] M. P. Leubner, "A nonextensive entropy approach to Kappa-distributions,"
       Astrophysics and Space Science, vol. 282, no. 3, pp. 573-579, 2002.
       doi:10.1029/2000JA000425
    """

    def __init__(self, params: isr.Parameters, particle: isr.Particle):
        """Initialize VDF parameters.

        Parameters
        ----------
        params : Parameters
            `Parameters` object with the parameters of the simulation.
        particle : Particle
            `Particle` object with the parameters of the particle.
        """
        self.params = params
        self.particle = particle
        self.normalize()

    def normalize(self) -> None:
        """Normalize the VDF."""
        self.v_th = np.sqrt(self.particle.temperature * const.k / self.particle.mass)
        self.A = (
            (np.pi * self.particle.kappa * self.v_th**2) ** (-3 / 2)
            * sps.gamma(self.particle.kappa)
            / sps.gamma(self.particle.kappa - 3 / 2)
        )

    def f_0(self) -> np.ndarray:
        """Define the phase space velocity distribution."""
        return self.A * (
            1 + self.particle.velocity_axis**2 / (self.particle.kappa * self.v_th**2)
        ) ** (-self.particle.kappa)


class VdfGaussShell(isr.Vdf):
    """Create an object that make Gauss shell distribution functions.

    This implementation need more testing and refining.
    """

    def __init__(self, params: isr.Parameters, particle: isr.Particle):
        """Initialize VDF parameters.

        Parameters
        ----------
        params : Parameters
            `Parameters` object with the parameters of the simulation.
        particle : Particle
            `Particle` object with the parameters of the particle.
        """
        self.params = params
        self.particle = particle
        temp_superthermal = getattr(self.particle, "temperature_superthermal", 90000)
        self.vth = np.sqrt(self.particle.temperature * const.k / self.particle.mass)
        self.r = (temp_superthermal * const.k / self.particle.mass) ** 0.5
        self.steep = 5
        self.f_M = isr.VdfMaxwell(self.params, self.particle)
        self.normalize()

    def normalize(self) -> None:
        """Normalize the VDF."""
        func = np.exp(
            -self.steep
            * (abs(self.particle.velocity_axis) - self.r) ** 2
            / (2 * self.particle.temperature * const.k / self.particle.mass)
        )
        f = func * self.particle.velocity_axis**2 * 4 * np.pi
        self.A = 1 / si.simps(f, self.particle.velocity_axis)
        ev = 0.5 * const.m_e * self.r**2 / const.eV
        print(f"Gauss shell at E = {round(ev, 2)} eV")

    def f_0(self) -> np.ndarray:
        """Define the phase space velocity distribution."""
        func = (
            self.A
            * np.exp(
                -self.steep
                * (abs(self.particle.velocity_axis) - self.r) ** 2
                / (2 * self.particle.temperature * const.k / self.particle.mass)
            )
            + 1e4 * self.f_M.f_0()
        )

        return func / (1e4 + 1)
